/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include "dixie.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <signal.h>
#include "packet.h"

extern int	debug;
extern int	resultlen;

static Packet	*pkt_queue;

/*
 * make_pkt - make a packet out of buffer b.  The packet is marked with
 * the current time so it can be expired by pkt_insert later.
 */

Packet *make_pkt(b, from, len, outgoing)
char			*b;
struct sockaddr_in	*from;
int			len;
char			outgoing;
{
	Packet	*p;

	if ((p = (Packet *) malloc(sizeof(Packet))) == NULLPKT) {
		perror("make_pkt");
		clean_exit(1);
	}

	p->pkt_time = time(0);
	if (from == 0)
		bzero(&p->pkt_from, sizeof(struct sockaddr_in));
	else
		p->pkt_from = *from;
	p->pkt_size = len;
	p->pkt_outgoing = outgoing;
	if ((p->pkt_data = (char *) malloc(len)) == NULL) {
		perror("make_pkt data");
		clean_exit(1);
	}
	bcopy(b, p->pkt_data, len);

	return(p);
}

/*
 * pkt_insert - insert a packet into the queue of outstanding packets.
 * as the packet is being inserted, the queue is scanned for old packets
 * which are discarded if they are found to be older than PKT_TIMEOUT.
 */

pkt_insert(p)
Packet	*p;
{
	Packet	*cur, *prev;
	long	curtime;
	int	rc;

	/* scan the queue for old packets */
	curtime = time(0);
	cur = pkt_queue;
	prev = NULLPKT;
	rc = PKT_OK;
	while ( cur != NULLPKT ) {
		if ( (! p->pkt_outgoing) && pkt_cmp(p, cur) == 0 )
			rc = PKT_DUPLICATE;

		if ( debug & 32 )
			printf("packet times (%d) (%d)\n", cur->pkt_time,
			    curtime);
		if ( curtime - cur->pkt_time > PKT_TIMEOUT ) {
			if ( debug & 32 ) {
				printf("packet timeout\n");
			}
			if ( prev == NULL ) {
				pkt_queue = cur->pkt_next;
				free(cur->pkt_data);
				free((char *) cur);
				cur = pkt_queue;
			} else {
				prev->pkt_next = cur->pkt_next;
				free(cur->pkt_data);
				free((char *) cur);
				cur = prev->pkt_next;
			}
		} else {
			prev = cur;
			cur = cur->pkt_next;
		}
	}

	if ( rc != PKT_OK ) {
		free(p->pkt_data);
		free((char *) p);
		return( rc );
	}

	/* insert this packet at the head of the queue */
	p->pkt_next = pkt_queue;
	pkt_queue = p;

	return( PKT_OK );
}

/*
 * pkt_cmp - compare two packets, return 0 if equal
 */

pkt_cmp(a, b)
Packet	*a;
Packet	*b;
{
	if ( bcmp((char *) &a->pkt_from, (char *) &b->pkt_from,
	    sizeof(struct sockaddr_in)) == 0 && bcmp(&a->pkt_data[DXI_H_ID],
	    &b->pkt_data[DXI_H_ID], sizeof(short)) == 0 )
		return(0);
	else
		return(1);
}

/*
 * pkt_remove - remove a packet from the queue of outstanding packets
 * in response to an ack.
 */

pkt_remove(p)
Packet			*p;
{
	Packet	*cur, *prev;

	cur = pkt_queue;
	prev = NULLPKT;
	while ( cur != NULLPKT ) {
		if ( pkt_cmp(cur, p) == 0 ) {	/* found it! */
			if ( prev == NULLPKT )
				pkt_queue = pkt_queue->pkt_next;
			else
				prev->pkt_next = cur->pkt_next;
			free(cur->pkt_data);
			free((char *) cur);

			return(PKT_FOUND);
		}

		prev = cur;
		cur = cur->pkt_next;
	}

	return(PKT_NOTFOUND);
}

/*
 * pkt_nack - arrange to have a particular packet (or piece of a packet)
 * resent.  The nack packet looks like this:
 *	header id sequence
 * where id is the id of the packet being nacked and sequence is the 
 * fragment number wanted.
 */

pkt_nack(b, from, rp, rseq, rid, rtotlen)
char			*b;
struct sockaddr_in	*from;
char			*rp;
short			*rseq;
short			*rid;
long			*rtotlen;
{
	short	id, curid;
	short	sequence;
	short	netshort;
	long	netlong;
	int	offset;
	long	totlen;
	int	pktlen;
	Packet	*cur;
	int	idcmp, fromcmp;

	*rp = '\0';
	/* extract the id and sequence number we're looking for */
	bcopy(&b[DXI_H_SIZE], (char *) &netshort, sizeof(short));
	id = ntohs(netshort);
	bcopy(&b[DXI_H_SIZE+sizeof(short)], (char *) &netshort, sizeof(short));
	sequence = ntohs(netshort);
	*rid = id;
	*rseq = sequence;
	if (debug & 16) printf("pkt_nack: id %d sequence %d\n", id, sequence);

	/* find the packet */
	cur = pkt_queue;
	while ( cur != NULLPKT ) {
		bcopy(&cur->pkt_data[DXI_H_ID], (char *)&netshort,
		    sizeof(short));
		curid = ntohs(netshort);
		idcmp = (id != curid );
		fromcmp = bcmp(from,&cur->pkt_from,sizeof(struct sockaddr_in));
		if (idcmp == 0 && fromcmp == 0 && cur->pkt_outgoing )
			break;

		cur = cur->pkt_next;
	}

	if (cur == NULLPKT) {
		if (debug) printf("packet not found\n");
		return(DXI_SEQNOTFOUND);
	}

	/* calculate fragment offset */
	offset = sequence * MAX_PKT_FRAG + DXI_H_SIZE;
	bcopy(&(cur->pkt_data[DXI_H_TOTLEN]), (char *) &netlong, sizeof(long));
	totlen = ntohl(netlong);
	*rtotlen = totlen;
	if (debug)
		printf("sequence %d offset %d totlen %d\n",sequence,offset,
		    totlen);
	if ( offset > cur->pkt_size ) {
		if ( debug )
			printf("preposterous sequence # (%d)\n", sequence);
		return(DXI_ERROR);
	}

	/* copy the fragment to the result buffer */
	pktlen = totlen - sequence * MAX_PKT_FRAG >= MAX_PKT_FRAG
	    ? MAX_PKT_FRAG : totlen - sequence * MAX_PKT_FRAG;
	bcopy(cur->pkt_data + offset, rp, pktlen);
	*(rp + pktlen) = '\0';

	return(cur->pkt_data[DXI_H_RC]);
}

/*
 * pkt_resend - resend an entire packet.  The packet should be in the
 * packet queue.  If so, it is copied into the result buffer.
 */

pkt_resend(b, from, r)
char			*b;
struct sockaddr_in	*from;
char			*r;
{
	short	id, curid;
	short	netshort;
	int	idcmp, fromcmp;
	Packet	*cur;

	bcopy(&b[DXI_H_ID], (char *) &netshort, sizeof(short));
	id = ntohs(netshort);

	/* find the packet */
	cur = pkt_queue;
	while ( cur != NULLPKT ) {
		bcopy(&cur->pkt_data[DXI_H_ID], (char *)&netshort,
		    sizeof(short));
		curid = ntohs(netshort);
		idcmp = (id != curid );
		fromcmp = bcmp(from,&cur->pkt_from,sizeof(struct sockaddr_in));
		if (idcmp == 0 && fromcmp == 0 && cur->pkt_outgoing)
			break;

		cur = cur->pkt_next;
	}

	if (cur == NULLPKT) {
		if (debug) printf("packet not found - retry operation\n");
		return(DXI_SEQNOTFOUND);
	}

	bcopy(&cur->pkt_data[DXI_H_SIZE], r, cur->pkt_size - DXI_H_SIZE);
	return(DXI_SUCCESS);
}
